package fr.jelora.teleinformationedf;

import java.io.File;
import java.io.FileInputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.hibernate.Query;
import org.hibernate.Session;

import fr.jelora.teleinformationedf.db.entity.ReleveTeleinformation;
import fr.jelora.teleinformationedf.db.hibernate.HibernateUtil;
import jssc.SerialPort;

public class TeleinformationEDF {
	
	final static Logger logger = Logger.getLogger(TeleinformationEDF.class);
	
	private String version = "0.1";
	private String dateVersion = "04/11/2017";
	
	private Session session;
	private String identifiantCompteur = null;
	private String serveurEnvoiReleveAdresseTemplate = null;
   
	public static void main( String[] args ){
		TeleinformationEDF teleinformationEDF = new TeleinformationEDF();
		teleinformationEDF.start();
    }
    
    public void start(){
    	try{
    		logger.info("Récupération consommation via Téléinformation EDF v" + version + " (" + dateVersion + ")");
    		
    		// Lecture configuration
			File fichierConfig = new File("config.properties");
			Properties config = new Properties();
			config.load(new FileInputStream(fichierConfig));
			String dbServeur = null;
			String dbNom = null;
			String dbIdentifiant = null;
			String dbMotDePasse = null;
			if (config.get("db.serveur") != null) {
				if (!config.get("db.serveur").toString().trim().isEmpty()) {
					dbServeur = config.get("db.serveur").toString().trim();
				}
			}
			if (config.get("db.nom") != null) {
				if (!config.get("db.nom").toString().trim().isEmpty()) {
					dbNom = config.get("db.nom").toString().trim();
				}
			}
			if (config.get("db.identifiant") != null) {
				if (!config.get("db.identifiant").toString().trim().isEmpty()) {
					dbIdentifiant = config.get("db.identifiant").toString().trim();
				}
			}
			if (config.get("db.motdepasse") != null) {
				if (!config.get("db.motdepasse").toString().trim().isEmpty()) {
					dbMotDePasse = config.get("db.motdepasse").toString().trim();
				}
			}
			if (config.get("serveurEnvoiReleve.adresseTemplate") != null) {
				if (!config.get("serveurEnvoiReleve.adresseTemplate").toString().trim().isEmpty()) {
					serveurEnvoiReleveAdresseTemplate = config.get("serveurEnvoiReleve.adresseTemplate").toString().trim();
				}
			}
			if (config.get("compteur.identifiant") != null) {
				if (!config.get("compteur.identifiant").toString().trim().isEmpty()) {
					identifiantCompteur = config.get("compteur.identifiant").toString().trim();
				}
			}
			
			// Intialisation base de données
			HibernateUtil hibernateUtil = new HibernateUtil();
			session = hibernateUtil.buildSessionFactory(dbServeur, dbNom, dbIdentifiant, dbMotDePasse).openSession();
    		
			// Initialisation port série
			SerialPort serial = new SerialPort("/dev/ttyAMA0");
			serial.openPort();
			serial.setParams(SerialPort.BAUDRATE_1200, SerialPort.DATABITS_7, SerialPort.STOPBITS_1, SerialPort.PARITY_EVEN);
			
			// Configuration du timer pour récupération de la téléinformation toutes les minutes
			Date date = new Date();
			Calendar calendar = Calendar.getInstance();
			calendar.setTimeInMillis(date.getTime() + 60*1000);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			date = calendar.getTime();
			SimpleDateFormat formatter = new SimpleDateFormat("HH:mm:ss");
			logger.info("Premier relevé à " + formatter.format(date));
			Timer timerReleve = new Timer();
			timerReleve.scheduleAtFixedRate(
				new TimerTask(){
					@Override
					public void run() {
						try{
							effectuerReleve(serial);
						}
						catch(Exception e){
							logger.error(e);
							for ( int i = 0 ; i < e.getStackTrace().length ; i++ ){
								logger.error("	" + e.getStackTrace()[i]);
							}
						}
					}
				}, 
				date, 
				60*1000
			);
			
			// Configuration du timer pour l'envoi des données sur un serveur toutes les minutes (+30 secondes par rapport aux relevés)
			calendar.set(Calendar.SECOND, 30);
			date = calendar.getTime();
			Timer timerEnvoiServeur = new Timer();
			timerEnvoiServeur.scheduleAtFixedRate(
				new TimerTask(){
					@Override
					public void run() {
						try{
							envoyerReleves();
						}
						catch(Exception e){
							logger.error(e);
							for ( int i = 0 ; i < e.getStackTrace().length ; i++ ){
								logger.error("	" + e.getStackTrace()[i]);
							}
						}
					}
				}, 
				date, 
				60*1000
			);
			
			
			
		}
		catch(Exception e){
			logger.error(e);
			for ( int i = 0 ; i < e.getStackTrace().length ; i++ ){
				logger.error("	" + e.getStackTrace()[i]);
			}
		}
    }
    
    private void envoyerReleves(){
    	
    	// Récupération des 50 premiers relevés
    	Query requete = session.createQuery("SELECT r FROM ReleveTeleinformation r ORDER BY r.idReleveTeleinformation");
    	requete.setMaxResults(50);
    	@SuppressWarnings("unchecked")
		List<ReleveTeleinformation> results = requete.list();
    	if ( results != null ){
    		if ( !results.isEmpty() ){
    			List<ReleveTeleinformation> relevesEnvoyesASupprimer = new ArrayList<ReleveTeleinformation>();
    			
    			// Envoi
    			String adresse;
    			String reponse;
    			URL obj;
    			HttpURLConnection con;
    			byte[] bytesReponse;
    			SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    			for ( ReleveTeleinformation releve : results ){
    				
    				try{
    					logger.debug("Envoi relevé du " + formatter.format(releve.getDateHeure()));
    					
    					adresse = serveurEnvoiReleveAdresseTemplate;
    					adresse = adresse.replace("{dateHeure}", this.encodeParamHttpGet(formatter.format(releve.getDateHeure())));
    					adresse = adresse.replace("{identifiantCompteur}", releve.getIdentifiantCompteur());
    					adresse = adresse.replace("{periodeTarifaireEnCours}", releve.getPeriodeTarifaireEnCours());
    					adresse = adresse.replace("{intensiteSouscrite}", releve.getIntensiteSouscrite() + "");
    					adresse = adresse.replace("{compteurHeuresCreuses}", releve.getCompteurHeuresCreuses() + "");
    					adresse = adresse.replace("{compteurHeuresPleines}", releve.getCompteurHeuresPleines() + "");
    					adresse = adresse.replace("{puissanceApparente}", releve.getPuissanceApparente() + "");
    					
    					obj = new URL(adresse);
    					con = (HttpURLConnection) obj.openConnection();
    					con.setConnectTimeout(2000);
    					con.setRequestMethod("GET");
    					con.setRequestProperty("User-Agent", "RaspberryPi-Teleinformation-EDF");
    					
    					if ( logger.isDebugEnabled() ){
    						bytesReponse = IOUtils.toByteArray(con.getInputStream());
    						reponse = new String(bytesReponse);
    						logger.debug("Réponse serveur: \n" + reponse);
    					}
    					
    					if ( con.getResponseCode() == 200 ){
    						relevesEnvoyesASupprimer.add(releve);
    					}
    				}
    				catch(Exception e){
    					logger.error("Envoi erreur");
    					logger.error(e);
    					for ( int i = 0 ; i < e.getStackTrace().length ; i++ ){
    						logger.error("	" + e.getStackTrace()[i]);
    					}
    					break;
    				}
    			}
    			if ( relevesEnvoyesASupprimer.size() == results.size() ){
    				logger.info("Envoi relevés OK (" + relevesEnvoyesASupprimer.size() + "/" + results.size() + ")");
    			}
    			else{
    				logger.error("Envoi partiel (" + relevesEnvoyesASupprimer.size() + "/" + results.size() + ")");
    			}
    			
    			// Suppression des relevés envoyés
    			for ( ReleveTeleinformation reveleASupprimer : relevesEnvoyesASupprimer ){
    				session.delete(reveleASupprimer);
    			}
    			session.flush();
    		}
    	}
    	
    }
    
    public String encodeParamHttpGet(String uri) throws Exception {
		if ( uri != null ){
			return URLEncoder.encode(uri, "utf8");
		}
		else{
			return null;
		}
	}
    
    private void effectuerReleve(SerialPort serial) throws Exception {
    	List<String> trame = this.recupererTrame(serial);
    	if ( !trame.isEmpty() ){
			try{
				if ( this.verifChecksumTrame(trame) ){
					String identifiantCompteur = this.valeurLigne(trame, "ADCO");
					String periodeTarifaireEnCours = this.valeurLigne(trame, "PTEC");
					if ( identifiantCompteur != null && periodeTarifaireEnCours != null ){
						if ( !identifiantCompteur.trim().isEmpty() && !periodeTarifaireEnCours.trim().isEmpty() ){
							if ( identifiantCompteur.equals(this.identifiantCompteur) ){
								ReleveTeleinformation releveTeleinformation = new ReleveTeleinformation();
								
								releveTeleinformation.setDateHeure(new Date());
								
								StringBuilder releveOriginalBuilder = new StringBuilder();
								for ( String ligne : trame ){
									if ( releveOriginalBuilder.length() > 0 ){
										releveOriginalBuilder.append("\n");
									}
									releveOriginalBuilder.append(ligne);
								}
								releveTeleinformation.setReleveOriginal(releveOriginalBuilder.toString());
								
								releveTeleinformation.setIdentifiantCompteur(identifiantCompteur);
								releveTeleinformation.setPeriodeTarifaireEnCours(periodeTarifaireEnCours);
								releveTeleinformation.setIntensiteSouscrite(Integer.parseInt(this.valeurLigne(trame, "ISOUSC")));
								releveTeleinformation.setCompteurHeuresCreuses(Integer.parseInt(this.valeurLigne(trame, "HCHC")));
								releveTeleinformation.setCompteurHeuresPleines(Integer.parseInt(this.valeurLigne(trame, "HCHP")));
								releveTeleinformation.setPuissanceApparente(Integer.parseInt(this.valeurLigne(trame, "PAPP")));
								
								session.persist(releveTeleinformation);
								session.flush();
								
								logger.info("Relevé OK: Compteur HP" + releveTeleinformation.getCompteurHeuresPleines() + "/HC" + releveTeleinformation.getCompteurHeuresCreuses() + " Puissance apparente " + releveTeleinformation.getPuissanceApparente() + "W");
							}
							else{
								logger.error("Relevé incorrect: identifiant compteur incorrect");
							}
						}
						else{
							logger.error("Relevé incorrect: identifiant compteur et/ou période tarifaire manquante(s)");
						}
					}
					else{
						logger.error("Relevé incorrect: identifiant compteur et/ou période tarifaire null");
					}
				}
				else{
					logger.error("Relevé incorrect: checksum trame incorrect");
					logger.error("  TRAME:");
					for ( String ligne : trame ){
						logger.error("	" + ligne + " --- CHECKSUM : " + this.verifChecksumLigne(ligne));
					}
				}
			}
			catch(Exception e){
				logger.error("Relevé incorrect");
				logger.error(e);
				for ( int i = 0 ; i < e.getStackTrace().length ; i++ ){
					logger.error("	" + e.getStackTrace()[i]);
				}
				logger.error("  TRAME:");
				for ( String ligne : trame ){
					logger.error("	" + ligne);
				}
			}
		}
		else{
			logger.warn("Relevé incorrect: Trame vide !");
		}
    }
    
    private boolean verifChecksumTrame(List<String> trame){
    	boolean result = false;
    	if ( !trame.isEmpty() ){
    		result = true;
    		for ( String ligne : trame ){
    			if ( !this.verifChecksumLigne(ligne) ){
    				result = false;
    			}
    		}
    	}
    	return result;
    }
    
    private boolean verifChecksumLigne(String ligne){
    	if ( ligne.length() > 2 ){
			String ligneContenu = ligne.substring(0, ligne.length()-2);
			char checksumLigne = ligne.charAt(ligne.length()-1);
			
			int sum = 0;
			for ( int i = 0 ; i < ligneContenu.length() ; i++ ){
				sum += ligneContenu.charAt(i);
			}
			sum = sum & 0x3F;
			sum += 0x20;
			
			if ( sum == checksumLigne ){
				return true;
			}
			else{
				return false;
			}
		}
		else{
			return false;
		}
	}
    
    private String valeurLigne(List<String> trame, String nomLigne){
    	String valeur = null;
    	if ( trame != null ){
    		for ( String ligne : trame ){
    			if ( ligne.startsWith(nomLigne) ){
    				String[] data = ligne.split(" ");
    				if ( data.length > 1 ){
    					valeur = data[1];
    				}
    				break;
    			}
    		}
    	}
    	return valeur;
    }
    
    private List<String> recupererTrame(SerialPort serial) throws Exception {
    	List<String> trame = new ArrayList<String>();
    	if ( serial != null ){
    		if ( serial.isOpened() ){
    			
    			boolean trameCommence = false;
    			boolean trameTerminee = false;
    			StringBuilder ligneBuilder = new StringBuilder();
    			String ligne = null;
    			byte[] buffer = null;
    			while ( !trameTerminee ){
    				buffer = serial.readBytes(1);
    				if ( buffer != null ){
    					ligneBuilder.append(new String(buffer));
    					if ( ligneBuilder.toString().endsWith("\n") ){
    						if ( ligneBuilder.toString().startsWith("ADCO ") && !trameCommence ){
    							trameCommence = true;
    						}
    						else if ( ligneBuilder.toString().startsWith("ADCO ") && trameCommence ){
    							trameCommence = false;
    							trameTerminee = true;
    						}
    						
    						if ( trameCommence ){
    							ligne = ligneBuilder.toString();
    							
    							// Vérification pour éviter d'effacer un checksum égal à un espace au moment du trim
    							if ( ligne.split(" ").length > 3 ){
    								ligne = ligne.trim() + "  ";
    							}
    							else{
    								ligne = ligne.trim();
    							}
    							trame.add(ligne);
    						}
    						ligneBuilder = new StringBuilder();
    					}
    				}
    				else{
    					Thread.sleep(10);
    				}
    			}
    			
    		}
    	}
    	return trame;
	}
}
